"""Generic support functions"""
import numpy as np
import pandas as pd
import math


def prop_table(y, pred, axis=0, round=2):
    """Makes a proportions table for a vector of binary predictions."""
    tab = pd.crosstab(y, pred)
    if axis == 1:
        tab = tab.transpose()
        out = tab / np.sum(tab, axis=0)
        out = out.transpose()
    else:
        out = tab / np.sum(tab, axis=0)
    if round is not None:
        out = np.round(out, round)
    return out


def onehot_matrix(y, sparse=False):
    """Converts a 1-D vector of class guesses to a one-hot matrix."""
    if not sparse:
        y_mat = np.zeros((y.shape[0], len(np.unique(y))))
        for row, col in enumerate(y):
            y_mat[row, col] = 1
    return y_mat


def max_probs(arr, maxes=None, axis=1):
    if maxes is None:
        maxes = np.argmax(arr, axis=axis)
    out = [arr[i, maxes[i]] for i in range(arr.shape[0])]
    return np.array(out)


def smash_log(x, B=10, d=0):
    """Logistic function with a little extra kick."""
    return 1 / (1 + np.exp(-x * B)) - d


def sparsify(col, 
             reshape=True, 
             return_df=True,
             long_names=False):
    """Makes a sparse array out of data frame of discrete variables."""
    levels = np.unique(col)
    out = np.array([col == level for level in levels],
                   dtype=np.uint8).transpose()
    if long_names:
        var = col.name + '.'
        levels = [var + level for level in levels]
    columns = [col.lower() for col in levels]
    if return_df:
        out = pd.DataFrame(out, columns=columns)
    return out


def row_sums(c, min=1):
    sums = np.sum(c, axis=1)
    return np.array(sums >= min, dtype=np.uint8)


def pair_sum(X, c, min=(1, 1)):
    a = row_sums(X[:, c[0]], min=min[0])
    b = row_sums(X[:, c[1]], min=min[1])
    return (a, b)


def combo_sum(ctup):
    sums = np.sum(ctup, axis=0)
    and_y = np.array(sums == 2, dtype=np.uint8)
    or_y = np.array(sums >= 1, dtype=np.uint8)
    out = np.concatenate([and_y.reshape(-1, 1),
                          or_y.reshape(-1, 1)],
                         axis=1)
    return out


def flatten(l):
    '''Flattens a list.'''
    return [item for sublist in l for item in sublist]


def unique_combo(c):
    """Determines if a combination of symptoms is unique."""
    if len(np.intersect1d(c[0], c[1])) == 0:
        return c
    else:
        return None


def pair_info(pair, link, col_list):
    """Turns lists of column strings and a prefix into a compound rule."""
    pair_cols = pair[1]
    pair_m = pair[2]
    m1 = pair_m[0]
    m2 = pair_m[1]
    n1 = len(pair_cols[0])
    n2 = len(pair_cols[1])
    
    # Making the string specifying the condition
    cnames = []
    for cols in pair_cols:
        cnames.append(' '.join([str(col_list[c]) 
                                     for c in cols]))
    out = [
        m1, cnames[0], link,
        m2, cnames[1], n1, n2
    ]
    return out


def zm_to_y(z, m, X):
    """Converts a variable choice vector, minimum count, and variables 
    to a binary guess vector.
    """
    return np.array(np.dot(X, z) >= m, dtype=np.uint8)


def zm_to_rule(z, m, cols, rule_num=1, return_df=True):
    """Converts a choice vector and minimum count to a rule string."""
    var_names = [cols[i] for i in np.where(z == 1)[0]]
    n = len(var_names)
    rule = ' '.join(var_names)
    m_col = 'm' + str(rule_num)
    n_col = 'n' + str(rule_num)
    rule_col = 'rule' + str(rule_num)
    if return_df:
        out = pd.DataFrame([m, rule, n]).transpose()
        out.columns = [m_col, rule_col, n_col]
    else:
        out = m, rule, n
    return out


def split_rule(rule, var_names):
    """Splits a single-string compound rule into two sub-rules."""
    link_dict = {'and': 1, 'or': 0}
    link_id = np.where([x not in var_names for x in rule])[0][0]
    link_val = link_dict[rule[link_id]]
    return rule[:link_id], rule[link_id + 1:], link_val
    

def rule_to_y(X, rule_df):
    """Converts a rule from DataFrame results format to a vector of guesses."""
    if rule_df.shape[0] > 1:
        rule_df = rule_df.to_frame().transpose()
    
    link_dict = {'and': 1, 'or': 0}
    cols1 = list(rule_df.rule1.values)[0].split()
    y1 = np.array(X[cols1].sum(1) >= rule_df.m1.values[0],
                  dtype=np.uint8)
    if ('n2' in rule_df.columns.values) and (rule_df.n2.values != 0):
        link_val = link_dict[rule_df.link.values[0]]
        cols2 = list(rule_df.rule2.values)[0].split()
        y2 = np.array(X[cols2].sum(1) >= rule_df.m2.values[0],
                      dtype=np.uint8)
        return np.array(y1 + y2 > link_val,
                        dtype=np.uint8)  
    else:
        return y1


def rule_to_str(rule_df, n):
    n_str = str(n)
    m = rule_df['m' + n_str].values[0]
    n = rule_df['n' + n_str].values[0]
    rule = rule_df['rule' + n_str].values[0]
    return 'At least ' + str(m) + ' of (' + rule + ')'


def rule_df_to_str(rule_df):
    out = rule_to_str(rule_df, 1)
    if 'n2' in rule_df.columns.values:
        if rule_df.n2 != 0:
            out += ' ' + rule_df.link + ' At least ' + rule_to_str(rule_df, 2)
    return out


def tab_to_vecs(tp, tn, fp, fn):
    tp = np.array([[1, 1]] * tp)
    tn = np.array([[0, 0]] * tn)
    fn = np.array([[1, 0]] * fn)
    fp = np.array([[0, 1]] * fp)
    good = [a for a in [tp, tn, fn, fp] if a.shape[0] != 0]
    return np.concatenate(good)



"""Various metrics for measuring classification performance."""
import numpy as np
import pandas as pd

from scipy.special import expit, erf
from sklearn.metrics import confusion_matrix
from multiprocessing import shared_memory

from .tools import smash_log, zm_to_y, pair_sum, combo_sum


fn_dict = {'j': ['sens', 'spec'],
           'f1': ['sens', 'ppv'],
           'mcc': ['j', 'mk']}


def brier_score(targets, guesses):
    """Calculates Brier score."""
    n_classes = len(np.unique(targets))
    assert n_classes > 1
    if n_classes == 2:
        bs = np.sum((guesses - targets)**2) / targets.shape[0]
    else:
        y = onehot_matrix(targets)
        row_diffs = np.diff((guesses, y), axis=0)[0]
        squared_diffs = row_diffs ** 2
        row_sums = np.sum(squared_diffs, axis=1) 
        bs = row_sums.mean()
    return bs


def mcc(y, y_, undef_val=0):
    """Calculates Matthews Correlation Coefficient."""
    prod = j(y, y_) * j(y_, y) * mk(y, y_) * mk(y_, y)
    return prod ** (1/4) if prod != 0 else undef_val


def f1(y, y_):
    """Alternative call for f_score()."""
    return f_score(y, y_, b=1)


def f_score(y, y_, b=1, undef_val=0):
    """Calculates F-score."""
    se = sens(y, y_)
    pv = ppv(y, y_)
    if se + pv != 0:
        return (1 + b**2) * (se * pv) / ((b**2 * pv) + se)
    else:
        return undef_val


def sens(y, y_):
    """Calculates sensitivity, or recall."""
    tp = np.sum((y ==1) & (y_ == 1))
    return tp / y.sum()


def spec(y, y_):
    """Calculates specificity, or 1 - FPR."""
    tn = np.sum((y == 0) & (y_ == 0))
    return tn / np.sum(y == 0)


def ppv(y, y_):
    """Calculates positive predictive value, or precision."""
    tp = np.sum((y == 1) & (y_ == 1))
    return tp / y_.sum()


def npv(y, y_):
    """Calculates negative predictive value."""
    tn = np.sum((y == 0) & (y_ == 0))
    return tn / np.sum(y_ == 0)


def sens_exp(z, xp, B=100):
    """Approximates sensitivity, or true positive rate, using the smash_log()
    function.
    """
    m = z[-1]
    z = z[:-1]
    return smash_log(np.dot(xp, z) - m, B=B).sum() / xp.shape[0]


def spec_exp(z, xn, B=100):
    """Approximates specificity, or 1 minus FPR, using the smash_log() 
    function.
    """
    m = z[-1]
    z = z[:-1]
    return 1 - smash_log(np.dot(xn, z) - m, B=B).sum() / xn.shape[0]


def mk(y, y_):
    """Calculates markedness, or PPV + NPV - 1."""
    return ppv(y, y_) + npv(y, y_) - 1


def j(y, y_, a=1, b=1):
    """Calculates Youden's J index from two binary vectors."""
    c = a + b
    a = a / c * 2
    b = b / c * 2
    sens = np.sum((y == 1) & (y_ == 1)) / y.sum()
    spec = np.sum((y == 0) & (y_ == 0)) / (len(y) - y.sum())
    return a*sens + b*spec - 1


def j_lin(z, m, X, y):
    """Calculates Youden's J index as a linear combination of a variable 
    choice vector, two variable matrices, and m."""
    z = np.round(z)
    y_ = zm_to_y(z, m, X)
    return j(y, y_)


def j_lin_comp(z_mat, m_vec, X, y):
    """Calculates Youden's J index from the N matrix and M vector specifying
    a given compound rule. Both must be binary.
    """
    guesses = np.array([zm_to_y(z_mat[:, i], m_vec[i], X)
                        for i in range(z_mat.shape[1])])
    y_ = np.array(np.sum(guesses, 0) > 0, dtype=np.uint8)
    stat = j(y, y_)
    return stat


def j_exp(z, xp, xn, a=1, b=1):
    """Calculates Youden's J index for a single m-of-n rule using the 
    parameters of a solved LP (z) and the smash_log() function.
    """
    m = z[-1]
    z = smash_log(z[:-1] - .5)
    tpr = smash_log(np.dot(xp, z) - m + .5).sum() / xp.shape[0]
    fpr = smash_log(np.dot(xn, z) - m + .5).sum() / xn.shape[0]
    return -1 * (a*tpr - b*fpr)


def j_exp_comp(z, xp, xn, c=2, a=1, b=1, th=0):
    """Calculates Youden's J index for a compound m-of-n rule using the
    parameters of a solved LP (z) and the smash_log() function.
    """
    # Setting things up
    s = xp.shape[1]
    m = z[-c:]
    z = z[:-c]
    z = z.reshape((s, c), order='F')
    z = smash_log(z - .5, B=15)
    
    # Penalizing bins where m > n
    nvals = z.sum(0)
    diffs = smash_log(nvals - m - .5)
    mn_penalty = th * (c - diffs.sum())
    
    # Now calculating the hits
    p_hits = smash_log(smash_log(np.dot(xp, z) - m + .5).sum(1) - .5).sum()
    n_hits = smash_log(smash_log(np.dot(xn, z) - m + .5).sum(1) - .5).sum()
    
    tpr = p_hits / xp.shape[0]
    fpr = n_hits / xn.shape[0] 
    weighted_j = a*tpr - b*fpr
    
    return -1 * weighted_j + mn_penalty


def pair_score(X, c, mc, y, metric):
    """Scores a pair of sets of columns."""
    ps = pair_sum(X, c, mc)
    cs = combo_sum(ps)
    and_scores = score_set(y, cs[:, 0], metric)
    or_scores = score_set(y, cs[:, 1], metric)
    return [and_scores, or_scores]


def score_set(y, y_, metric, return_df=False):
    """Provides a small set of scores for a set of predictions."""
    fn1 = globals()[fn_dict[metric][0]]
    fn2 = globals()[fn_dict[metric][1]]
    fn3 = globals()[metric]
    score_fns = [fn1, fn2, fn3]
    scores = [fn(y, y_) for fn in score_fns]
    if return_df:
        out = pd.DataFrame(scores).transpose()
        out.columns = fn_dict[metric] + [metric]
        return out
    else:
        return scores


def shared_mem_score(c, m, xshape, metric='j'):
    """Pulls X and y from shared memory and then scores a combination."""
    X_buf = shared_memory.SharedMemory(name='predictors')
    y_buf = shared_memory.SharedMemory(name='outcome')
    X = np.ndarray(xshape, dtype=np.uint8, buffer=X_buf.buf)
    y = np.ndarray((xshape[0],), dtype=np.uint8, buffer=y_buf.buf)
    y_ = np.array(X[:, c].sum(1) > m, dtype=np.uint8)
    scores = score_set(y, y_, metric)
    X_buf.close()
    y_buf.close()
    return scores


def risk_ratio(y, pred, round=2):
    """Calculates the risk ratio."""
    props = np.array(prop_table(y, pred, round=None))
    rr = props[1, 1] / props[1, 0]
    if round is not None:
        rr = np.round(rr, round)
    return rr


def odds_ratio(y, pred, round=2):
    """Calculates the odds ratio."""
    tab = np.array(pd.crosstab(y, pred))
    OR = (tab[0, 0]*tab[1, 1]) / (tab[1, 0]*tab[0, 1])
    if round is not None:
        OR = np.round(OR, round)
    return OR


def clf_metrics(true, 
                pred,
                average='weighted',
                preds_are_probs=False,
                cutpoint=0.5,
                mod_name=None,
                round=4,
                round_pval=False,
                mcnemar=False,
                argmax_axis=1,
                undef_val=0):
    '''Runs basic diagnostic stats on binary (only) predictions'''
    # Converting pd.Series to np.array
    stype = type(pd.Series([0]))
    if type(pred) == stype:
        pred = pred.values
    if type(true) == stype:
        true = true.values
    
    # Optional exit for doing averages with multiclass/label inputs
    if len(np.unique(true)) > 2:
        # Getting binary metrics for each set of results
        codes = np.unique(true)
        
        # Argmaxing for when we have probabilities
        if preds_are_probs:
            auc = roc_auc_score(true,
                                pred,
                                average=average,
                                multi_class='ovr')
            brier = brier_score(true, pred)
            pred = np.argmax(pred, axis=argmax_axis)
        
        # Making lists of the binary predictions (OVR)    
        y = [np.array([doc == code for doc in true], dtype=np.uint8)
             for code in codes]
        y_ = [np.array([doc == code for doc in pred], dtype=np.uint8)
              for code in codes]
        
        # Getting the stats for each set of binary predictions
        stats = [clf_metrics(y[i], y_[i], round=16) for i in range(len(y))]
        stats = pd.concat(stats, axis=0)
        stats.fillna(0, inplace=True)
        cols = stats.columns.values
        
        # Calculating the averaged metrics
        if average == 'weighted':
            weighted = np.average(stats, 
                                  weights=stats.true_prev,
                                  axis=0)
            out = pd.DataFrame(weighted).transpose()
            out.columns = cols
        elif average == 'macro':
            out = pd.DataFrame(stats.mean()).transpose()
        elif average == 'micro':
            out = clf_metrics(np.concatenate(y),
                              np.concatenate(y_))
        
        # Adding AUC and AP for when we have probabilities
        if preds_are_probs:
            out.auc = auc
            out.brier = brier
        
        # Rounding things off
        out = out.round(round)
        count_cols = [
                      'tp', 'fp', 'tn', 'fn', 'true_prev',
                      'pred_prev', 'prev_diff'
        ]
        out[count_cols] = out[count_cols].round()
        
        if mod_name is not None:
            out['model'] = mod_name
        
        return out
    
    # Thresholding the probabilities, if provided
    if preds_are_probs:
        auc = roc_auc_score(true, pred)
        brier = brier_score(true, pred)
        ap = average_precision_score(true, pred)
        pred = threshold(pred, cutpoint)
    else:
        brier = np.round(brier_score(true, pred), round)
    
    # Constructing the 2x2 table
    confmat = confusion_matrix(true, pred)
    tp = confmat[1, 1]
    fp = confmat[0, 1]
    tn = confmat[0, 0]
    fn = confmat[1, 0]
    
    # Calculating the main binary metrics
    ppv = np.round(tp / (tp + fp), round) if tp + fp > 0 else undef_val
    sens = np.round(tp / (tp + fn), round) if tp + fn > 0 else undef_val
    spec = np.round(tn / (tn + fp), round) if tn + fp > 0 else undef_val
    npv = np.round(tn / (tn + fn), round) if tn + fn > 0 else undef_val
    if sens + ppv != 0:
        f1 = np.round(2*(sens*ppv) / (sens+ppv), round)  
    else:
        f1 = undef_val
    
    # Calculating the Matthews correlation coefficient
    mcc_num = ((tp * tn) - (fp * fn))
    mcc_denom = np.sqrt(((tp+fp)*(tp+fn)*(tn+fp)*(tn+fn)))
    mcc = mcc_num / mcc_denom if mcc_denom != 0 else undef_val
    
    # Calculating Youden's J and the Brier score
    j = sens + spec - 1
    
    # Rolling everything so far into a dataframe
    outmat = np.array([tp, fp, tn, fn,
                       sens, spec, ppv,
                       npv, j, f1, mcc, brier]).reshape(-1, 1)
    out = pd.DataFrame(outmat.transpose(),
                       columns=['tp', 'fp', 'tn', 
                                'fn', 'sens', 'spec', 'ppv',
                                'npv', 'j', 'f1', 'mcc', 'brier'])
    
    # Optionally tacking on stats from the raw probabilities
    if preds_are_probs:
        out['auc'] = auc
        out['ap'] = ap
    else:
        out['auc'] = 0
        out['ap'] = 0
    
    # Calculating some additional measures based on positive calls
    true_prev = int(np.sum(true == 1))
    pred_prev = int(np.sum(pred == 1))
    abs_diff = (true_prev - pred_prev) * -1
    rel_diff = np.round(abs_diff / true_prev, round)
    if mcnemar:
        pval = mcnemar_test(true, pred).pval[0]
        if round_pval:
            pval = np.round(pval, round)
    count_outmat = np.array([true_prev, pred_prev, abs_diff, 
                             rel_diff]).reshape(-1, 1)
    count_out = pd.DataFrame(count_outmat.transpose(),
                             columns=['true_prev', 'pred_prev', 
                                      'prev_diff', 'rel_prev_diff'])
    out = pd.concat([out, count_out], axis=1)
    
    # Optionally dropping the mcnemar p-val
    if mcnemar:
        out['mcnemar'] = pval
    
    # And finally tacking on the model name
    if mod_name is not None:
        stat_cols = list(out.columns.values)
        out['model'] = mod_name
        out = out[['model'] + stat_cols]
    
    return out




import pandas as pd
import numpy as np
import os

from sklearn.metrics import confusion_matrix
from sklearn.metrics import roc_auc_score, average_precision_score
from scipy.stats import binom, chi2, norm, percentileofscore
from multiprocessing import Pool
from copy import deepcopy

from .tools import *
from .metrics import *


def threshold(probs, cutoff=.5):
    return np.array(probs >= cutoff).astype(np.uint8)


def mcnemar_test(true, pred, cc=True):
    cm = confusion_matrix(true, pred)
    b = int(cm[0, 1])
    c = int(cm[1, 0])
    if cc:
        stat = (abs(b - c) - 1)**2 / (b + c)
    else:
        stat = (b - c)**2 / (b + c)
    p = 1 - chi2(df=1).cdf(stat)
    outmat = np.array([b, c, stat, p]).reshape(-1, 1)
    out = pd.DataFrame(outmat.transpose(),
                       columns=['b', 'c', 'stat', 'pval'])
    return out


def jackknife_metrics(targets, 
                      guesses, 
                      average='weighted'):
    # Replicates of the dataset with one row missing from each
    rows = np.array(list(range(targets.shape[0])))
    j_rows = [np.delete(rows, row) for row in rows]

    # using a pool to get the metrics across each
    scores = [clf_metrics(targets[idx],
                          guesses[idx],
                          average=average) for idx in j_rows]
    scores = pd.concat(scores, axis=0)
    means = scores.mean()
    
    return scores, means


def invert_BCA(q, b, acc, lower=True):
    if np.isnan(q):
        return q
    if np.any(q >= 1):
        q /= 100
    z = norm.ppf(q)
    numer = z - 2*b - z*b*acc - b**2*acc
    denom = 1 + b*acc + z*acc
    if lower:
        return 2 * norm.cdf(numer / denom)
    else:
        return 2 * (1 - norm.cdf(numer / denom))


def BCA_pval(bca, null=0.0):
    scores = bca.scores
    n_cols = bca.scores.shape[1]
    
    # Making the vector of null values
    if type(null) == type(0.0):
        null = np.array([null] * n_cols)
    
    # Figuring out which nulls exist in the bootstrap data
    good_nulls = []
    for i in range(n_cols):
        col = scores.iloc[:, i]
        good_nulls.append(col.min() <= null[i] <= col.max())
    
    # Getting the percentile for each null
    null_qs = np.array([percentileofscore(bca.scores.iloc[:, i], null[i]) 
               if good_nulls[i] else np.nan for i in range(n_cols)]) / 100
    
    # Figuring out whether to look at the lower or upper quantile
    lower = [q <= .5 if not np.isnan(q) else q for q in null_qs]
    
    # Getting the p-value associated with each null percentile
    pvals = np.array([invert_BCA(null_qs[i],
                                 bca.b[i],
                                 bca.acc[i],
                                 lower=lower[i])
                      for i in range(n_cols)])
    
    return pvals, good_nulls, null_qs, lower


def boot_stat_cis(stat,
                  jacks,
                  boots,
                  a=0.05,
                  exp=False,
                  method="bca",
                  interpolation="nearest",
                  transpose=True,
                  outcome_axis=1,
                  stat_axis=2):
    # Renaming because I'm lazy
    j = jacks
    n = len(boots)
    
    # Calculating the confidence intervals
    lower = (a / 2) * 100
    upper = 100 - lower

    # Making sure a valid method was chosen
    methods = ["pct", "diff", "bca"]
    assert method in methods, "Method must be pct, diff, or bca."

    # Calculating the CIs with method #1: the percentiles of the
    # bootstrapped statistics
    if method == "pct":
        cis = np.nanpercentile(boots,
                               q=(lower, upper),
                               interpolation=interpolation,
                               axis=0)
        cis = pd.DataFrame(cis.transpose(),
                           columns=["lower", "upper"],
                           index=colnames)

    # Or with method #2: the percentiles of the difference between the
    # obesrved statistics and the bootstrapped statistics
    elif method == "diff":
        diffs = stat - boots
        percents = np.nanpercentile(diffs,
                                    q=(lower, upper),
                                    interpolation=interpolation,
                                    axis=0)
        lower_bound = pd.Series(stat_vals + percents[0])
        upper_bound = pd.Series(stat_vals + percents[1])
        cis = pd.concat([lower_bound, upper_bound], axis=1)
        cis = cis.set_index(stat.index)

    # Or with method #3: the bias-corrected and accelerated bootstrap
    elif method == "bca":
        # Calculating the bias-correction factor
        n_less = np.sum(boots < stat, axis=0)
        p_less = n_less / n
        z0 = norm.ppf(p_less)

        # Fixing infs in z0
        z0[np.where(np.isinf(z0))[0]] = 0.0

        # Estiamating the acceleration factor
        diffs = j[1] - j[0]
        numer = np.sum(np.power(diffs, 3))
        denom = 6 * np.power(np.sum(np.power(diffs, 2)), 3/2)

        # Getting rid of 0s in the denominator
        zeros = np.where(denom == 0)[0]
        for z in zeros:
            denom[z] += 1e-6

        # Finishing up the acceleration parameter
        acc = numer / denom

        # Calculating the bounds for the confidence intervals
        zl = norm.ppf(a / 2)
        zu = norm.ppf(1 - (a / 2))
        lterm = (z0 + zl) / (1 - acc * (z0 + zl))
        uterm = (z0 + zu) / (1 - acc * (z0 + zu))
        ql = norm.cdf(z0 + lterm) * 100
        qu = norm.cdf(z0 + uterm) * 100

        # Returning the CIs based on the adjusted quantiles;
        # I know this code is hideous
        if len(boots.shape) > 2:
            n_outcomes = range(boots.shape[outcome_axis])
            n_vars = range(boots.shape[stat_axis])
            cis = np.array([
                [np.nanpercentile(boots[:, i, j],
                                  q =(ql[i][j], 
                                      qu[i][j]),
                                  axis=0) 
                                  for i in n_outcomes]
                for j in n_vars
            ])
        else:
            n_stats = range(len(ql))
            cis = np.array([
                np.nanpercentile(boots[:, i],
                                 q=(ql[i], qu[i]),
                                 interpolation=interpolation,
                                 axis=0) 
                for i in n_stats])
        
        # Optional exponentiation for log-link models
        if exp:
            cis = np.exp(cis)
        
        # Optional transposition
        if transpose:
            cis = cis.transpose()

    return cis


# Calculates bootstrap confidence intervals for an estimator
class boot_cis:
    def __init__(
        self,
        targets,
        guesses,
        n=100,
        a=0.05,
        group=None,
        method="bca",
        interpolation="nearest",
        average='weighted',
        mcnemar=False,
        seed=10221983,
        undef_val=0):
        # Converting everything to NumPy arrays, just in case
        stype = type(pd.Series([0]))
        if type(targets) == stype:
            targets = targets.values
        if type(guesses) == stype:
            guesses = guesses.values

        # Getting the point estimates
        stat = clf_metrics(targets,
                           guesses,
                           average=average,
                           mcnemar=mcnemar,
                           undef_val=undef_val).transpose()

        # Pulling out the column names to pass to the bootstrap dataframes
        colnames = list(stat.index.values)

        # Making an empty holder for the output
        scores = pd.DataFrame(np.zeros(shape=(n, stat.shape[0])),
                              columns=colnames)

        # Setting the seed
        if seed is None:
            seed = np.random.randint(0, 1e6, 1)
        np.random.seed(seed)
        seeds = np.random.randint(0, 1e6, n)

        # Generating the bootstrap samples and metrics
        boots = [boot_sample(targets, 
                             by=group, 
                             seed=seed) for seed in seeds]
        scores = [clf_metrics(targets[b], 
                              guesses[b], 
                              average=average,
                              undef_val=undef_val) for b in boots]
        scores = pd.concat(scores, axis=0)
        self.scores = scores

        # Calculating the confidence intervals
        lower = (a / 2) * 100
        upper = 100 - lower

        # Making sure a valid method was chosen
        methods = ["pct", "diff", "bca"]
        assert method in methods, "Method must be pct, diff, or bca."

        # Calculating the CIs with method #1: the percentiles of the
        # bootstrapped statistics
        if method == "pct":
            cis = np.nanpercentile(scores,
                                   q=(lower, upper),
                                   interpolation=interpolation,
                                   axis=0)
            cis = pd.DataFrame(cis.transpose(),
                               columns=["lower", "upper"],
                               index=colnames)

        # Or with method #2: the percentiles of the difference between the
        # obesrved statistics and the bootstrapped statistics
        elif method == "diff":
            stat_vals = stat.transpose().values.ravel()
            diffs = stat_vals - scores
            percents = np.nanpercentile(diffs,
                                        q=(lower, upper),
                                        interpolation=interpolation,
                                        axis=0)
            lower_bound = pd.Series(stat_vals + percents[0])
            upper_bound = pd.Series(stat_vals + percents[1])
            cis = pd.concat([lower_bound, upper_bound], axis=1)
            cis = cis.set_index(stat.index)

        # Or with method #3: the bias-corrected and accelerated bootstrap
        elif method == "bca":
            # Calculating the bias-correction factor
            stat_vals = stat.transpose().values.ravel()
            n_less = np.sum(scores < stat_vals, axis=0)
            p_less = n_less / n
            z0 = norm.ppf(p_less)

            # Fixing infs in z0
            z0[np.where(np.isinf(z0))[0]] = 0.0

            # Estiamating the acceleration factor
            j = jackknife_metrics(targets, guesses)
            diffs = j[1] - j[0]
            numer = np.sum(np.power(diffs, 3))
            denom = 6 * np.power(np.sum(np.power(diffs, 2)), 3 / 2)

            # Getting rid of 0s in the denominator
            zeros = np.where(denom == 0)[0]
            for z in zeros:
                denom[z] += 1e-6

            # Finishing up the acceleration parameter
            acc = numer / denom
            self.jack = j

            # Calculating the bounds for the confidence intervals
            zl = norm.ppf(a / 2)
            zu = norm.ppf(1 - (a / 2))
            lterm = (z0 + zl) / (1 - acc * (z0 + zl))
            uterm = (z0 + zu) / (1 - acc * (z0 + zu))
            ql = norm.cdf(z0 + lterm) * 100
            qu = norm.cdf(z0 + uterm) * 100
            
            # Passing things back to the class
            self.acc = acc.values
            self.b = z0
            self.ql = ql
            self.qu = qu

            # Returning the CIs based on the adjusted quintiles
            cis = [
                np.nanpercentile(
                    scores.iloc[:, i],
                    q=(ql[i], qu[i]),
                    interpolation=interpolation,
                    axis=0,
                ) for i in range(len(ql))
            ]
            cis = pd.DataFrame(cis, 
                               columns=["lower", "upper"], 
                               index=colnames)

        # Putting the stats with the lower and upper estimates
        cis = pd.concat([stat, cis], axis=1)
        cis.columns = ["stat", "lower", "upper"]

        # Passing the results back up to the class
        self.cis = cis
        self.scores = scores

        return


def average_pvals(p_vals, 
                  w=None, 
                  method='harmonic',
                  smooth=True,
                  smooth_val=1e-7):
    if smooth:
        p = p_vals + smooth_val
    else:
        p = deepcopy(p_vals)
    if method == 'harmonic':
        if w is None:
            w = np.repeat(1 / len(p), len(p))
        p_avg = 1 / np.sum(w / p)
    elif method == 'fisher':
        stat = -2 * np.sum(np.log(p))
        p_avg = 1 - chi2(df=1).cdf(stat)
    return p_avg


def jackknife_sample(X):
    rows = np.array(list(range(X.shape[0])))
    j_rows = [np.delete(rows, row) for row in rows]
    return j_rows


def boot_sample(df,
                by=None,
                size=None,
                seed=None,
                return_df=False):
    
    # Setting the random states for the samples
    if seed is None:
        seed = np.random.randint(1, 1e6, 1)[0]
    np.random.seed(seed)
    
    # Getting the sample size
    if size is None:
        size = df.shape[0]
    
    # Sampling across groups, if group is unspecified
    if by is None:
        np.random.seed(seed)
        idx = range(size)
        boot = np.random.choice(idx,
                                size=size,
                                replace=True)
    
    # Sampling by group, if group has been specified
    else:
        levels = np.unique(by)
        n_levels = len(levels)
        level_ids = [np.where(by == level)[0]
                     for level in levels]
        boot = [np.random.choice(ids, size=len(ids), replace=True)
                for ids in level_ids]
        boot = np.concatenate(boot).ravel()
    
    if not return_df:
        return boot
    else:
        return df.iloc[boot, :]
    

class diff_boot_cis:
    def __init__(self,
                 ref,
                 comp,
                 a=0.05,
                 abs_diff=False,
                 method='bca',
                 interpolation='nearest'):
        # Quick check for a valid estimation method
        methods = ['pct', 'diff', 'bca']
        assert method in methods, 'Method must be pct, diff, or bca.'
        
        # Pulling out the original estiamtes
        ref_stat = pd.Series(ref.cis.stat.drop('true_prev').values)
        ref_scores = ref.scores.drop('true_prev', axis=1)
        comp_stat = pd.Series(comp.cis.stat.drop('true_prev').values)
        comp_scores = comp.scores.drop('true_prev', axis=1)
        
        # Optionally Reversing the order of comparison
        diff_scores = comp_scores - ref_scores
        diff_stat = comp_stat - ref_stat
            
        # Setting the quantiles to retrieve
        lower = (a / 2) * 100
        upper = 100 - lower
        
        # Calculating the percentiles 
        if method == 'pct':
            cis = np.nanpercentile(diff_scores,
                                   q=(lower, upper),
                                   interpolation=interpolation,
                                   axis=0)
            cis = pd.DataFrame(cis.transpose())
        
        elif method == 'diff':
            diffs = diff_stat.values.reshape(1, -1) - diff_scores
            percents = np.nanpercentile(diffs,
                                        q=(lower, upper),
                                        interpolation=interpolation,
                                        axis=0)
            lower_bound = pd.Series(diff_stat + percents[0])
            upper_bound = pd.Series(diff_stat + percents[1])
            cis = pd.concat([lower_bound, upper_bound], axis=1)
        
        elif method == 'bca':
            # Removing true prevalence from consideration to avoid NaNs
            ref_j_means = ref.jack[1].drop('true_prev')
            ref_j_scores = ref.jack[0].drop('true_prev', axis=1)
            comp_j_means = comp.jack[1].drop('true_prev')
            comp_j_scores = comp.jack[0].drop('true_prev', axis=1)
            
            # Calculating the bias-correction factor
            n = ref.scores.shape[0]
            stat_vals = diff_stat.transpose().values.ravel()
            n_less = np.sum(diff_scores < stat_vals, axis=0)
            p_less = n_less / n
            z0 = norm.ppf(p_less)
            
            # Fixing infs in z0
            z0[np.where(np.isinf(z0))[0]] = 0.0
            
            # Estiamating the acceleration factor
            j_means = comp_j_means - ref_j_means
            j_scores = comp_j_scores - ref_j_scores
            diffs = j_means - j_scores
            numer = np.sum(np.power(diffs, 3))
            denom = 6 * np.power(np.sum(np.power(diffs, 2)), 3/2)
            
            # Getting rid of 0s in the denominator
            zeros = np.where(denom == 0)[0]
            for z in zeros:
                denom[z] += 1e-6
            
            acc = numer / denom
            
            # Calculating the bounds for the confidence intervals
            zl = norm.ppf(a / 2)
            zu = norm.ppf(1 - (a/2))
            lterm = (z0 + zl) / (1 - acc*(z0 + zl))
            uterm = (z0 + zu) / (1 - acc*(z0 + zu))
            ql = norm.cdf(z0 + lterm) * 100
            qu = norm.cdf(z0 + uterm) * 100
                                    
            # Returning the CIs based on the adjusted quantiles
            cis = [np.nanpercentile(diff_scores.iloc[:, i], 
                                    q=(ql[i], qu[i]),
                                    interpolation=interpolation,
                                    axis=0) 
                   for i in range(len(ql))]
            cis = pd.DataFrame(cis, columns=['lower', 'upper'])
                    
        cis = pd.concat([ref_stat, comp_stat, diff_stat, cis], 
                        axis=1)
        cis = cis.set_index(ref_scores.columns.values)
        cis.columns = ['ref', 'comp', 'd', 
                       'lower', 'upper']
        
        # Passing stuff back up to return
        self.cis = cis
        self.scores = diff_scores
        self.b = z0
        self.acc = acc
        
        return


def grid_metrics(targets,
                 guesses,
                 step=.01,
                 min=0.0,
                 max=1.0,
                 by='f1',
                 average='binary',
                 counts=True):
    cutoffs = np.arange(min, max, step)
    if len((guesses.shape)) == 2:
        if guesses.shape[1] == 1:
            guesses = guesses.flatten()
        else:
            guesses = guesses[:, 1]
    if average == 'binary':
        scores = []
        for i, cutoff in enumerate(cutoffs):
            threshed = threshold(guesses, cutoff)
            stats = clf_metrics(targets, threshed)
            stats['cutoff'] = pd.Series(cutoff)
            scores.append(stats)
    
    return pd.concat(scores, axis=0)


def merge_cis(c, round=4, mod_name=''):
    str_cis = c.round(round).astype(str)
    str_paste = pd.DataFrame(str_cis.stat + ' (' + str_cis.lower + 
                                 ', ' + str_cis.upper + ')',
                                 columns=[mod_name]).transpose()
    return str_paste


def merge_ci_list(l, mod_names=None, round=4):
    if type(l[0] != type(pd.DataFrame())):
        l = [c.cis for c in l]
    if mod_names is not None:
        merged_cis = [merge_cis(l[i], round, mod_names[i])
                      for i in range(len(l))]
    else:
        merged_cis = [merge_cis(c, round=round) for c in l]
    
    return pd.concat(merged_cis, axis=0)




"""A setuptools module for kudos.
See:
https://packaging.python.org/en/latest/distributing.html
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
#with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    #long_description = f.read()

setup(
    name='kudos',
    version='0.0.1',
    description='Tools for optimizing public health case definitions',
    #long_description=long_description,
    #long_description_content_type='text/markdown',
    url='https://github.com/scotthlee/kudos',
    author='Scott Lee',
    author_email='yle4@cdc.gov',
    license='Apache 2.0',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    keywords='combinatorial optimization',
    packages=['kudos'],
    install_requires=[
        'matplotlib', 'numpy', 'ortools',
        'pandas', 'scikit-learn', 'scipy',
        'seaborn', 'tqdm'
    ],
)


"""Wrapper classes for different optimization functions."""
import pandas as pd
import numpy as np
import scipy as sp
import seaborn as sns
import sklearn
import math
import tqdm
import ortools

from multiprocessing import Pool, Array, shared_memory
from copy import deepcopy
from matplotlib import pyplot as plt
from ortools.linear_solver import pywraplp
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier, GradientBoostingClassifier
from matplotlib import pyplot as plt
from itertools import combinations, permutations, islice

from . import metrics, tools


metric_dict = {
    'j': {
        'fn1': 'sens',
        'fn2': 'spec',
        'xlab': '1 - Specificity',
        'ylab': 'Sensitivity'
    },
    'f1': {
        'fn1': 'sens',
        'fn2': 'ppv',
        'xlab': 'PPV',
        'ylab': 'Sensitivity'
    },
    'mcc': {
        'fn1': 'j',
        'fn2': 'mk',
        'xlab': 'Markedness',
        'ylab': 'Informedness'
    }
}



class FeaturePruner:
    """A wrapper for sklearn models that can be used to whittle down the 
    feature space for large problems.
    """
    def __init__(self,
                 model_type='rf',
                 n_jobs=-1, 
                 n_estimators=1000,
                 l1_ratio=0.5,
                 other_args=None):
        """Initializes the pruner.
        
        Parameters
        ----------
        model_type : str, default='rf'
          The kind of model to use as the pruner. Options are a random forest 
          ('rf'), a gradient boosting classifier ('gbc'), and a logistic
          regression with L1 ('l1'), L2 ('l2'), and L1+L2 ('elasticnet') 
          penalties.
        n_jobs : int, default=-1
          n_jobs parameter to pass to sklearn models. -1 uses all processes.
        n_estimators : int, default=1000
          n_estimators parameter to pass to sklearn.ensemble models.
        l1_ratio : float in (0, 1), default=0.5
          Blending parameter for the L!/L2 penalties in elasticnet.
        other_args : dict, default=None
          Dictionary of other args for the base model.
        
        Returns
        ----------
        A FeaturePruner object with a base model ready to fit.
        """
        tree_dict = {'rf': 'RandomForestClassifier',
                    'gbc': 'GradientBoostingClassifier'}
        self.tree_mods = ['rf', 'gbc']
        if model_type in self.tree_mods:
            args = {'n_estimators': n_estimators,
                    'n_jobs': n_jobs}
            if other_args:
                args = {**args, **other_args}
            mod_fn = getattr(sklearn.ensemble, 
                             tree_dict[model_type])
            self.mod = mod_fn(**args)
        else:
            args = {'penalty': model_type,
                    'solver': 'saga',
                    'l1_ratio': l1_ratio}
            if other_args:
                args = {**args, **other_args}
            self.mod = LogisticRegression(**args)
        
        self.mod_type = model_type
        
    def fit(self, X, y, 
            factor=0.5,
            return_x=False, 
            other_args=None):
        """Fits the FeaturePruner to a dataset.
        
        Parameters
        ----------
        X : pd.DataFrame
          The array of predictors.
        y : array-like
          The array of targets for prediction.
        factor : float or int, default=0.5
          Either the proportion or number of original features to keep.
        return_x : bool, default=False
          Whether to return the pruned predictors after fitting.
        other_args : dict, default=None
          A dict of other args to pass to the base model fit() function.
        
        Returns
        ----------
        None or a pd.DataFrame of the pruned predictors.
        """
        self.factor = factor
        self.var_names = X.columns.values
        if self.factor < 1:
            top_n = int(self.factor * X.shape[1])
        else:
            top_n = self.factor
        if other_args:
            self.mod.fit(X, y, **kwargs)
        else:
            self.mod.fit(X, y)
        if self.mod_type in self.tree_mods:
            imps = self.mod.feature_importances_
        else:
            imps = self.mod.coef_[0]
        
        sorted_imps = np.argsort(imps)[::-1]
        self.top_feature_ids = sorted_imps[:top_n]
        self.sorted_var_names = [self.var_names[f] for f in sorted_imps]  
        self.top_var_names = [self.var_names[f] for f in self.top_feature_ids]  
        self.X= X.iloc[:, self.top_feature_ids]
        if return_x:
            self.X
    
    def predict(self, X):
        """Returns a pruned version of a dataset.
        
        Parameters
        ----------
        X : pd.DataFrame or np.array
        
        Returns
        ----------
        """
        if type(X) == type(pd.DataFrame([0])):
            return X[self.top_var_names]
        else:
            return X[:, self.top_feature_ids]



class FullEnumeration:
    """A brute-force combinatorial optimizer. Hulk smash!"""
    def __init__(self, n_jobs=None, share_memory=False):
        self.n_jobs = n_jobs
        self.share_memory = share_memory
        return
    
    def fit(self, X, y,
            max_n=5,
            metric='j',
            compound=False,
            use_reverse=False,
            top_n=None,
            chunksize=1000,
            batch_keep_n=1000,
            write_full=False,
            csv_name='fe_results.csv',
            prune=True,
            tol=.05,
            progress_bars=True,
            verbose=False):
        """Fits the optimizer.
        
        Parameters
        ----------
        max_n : int, default=5
            The maximum allowable combination size.
        metric : str, default='j'
            The classification metric to be optimized. Options are 'j', 'f1',
            or 'mcc'.
        compound : bool, default=False
            Whether to search compound combinations. Performance will \
            probably be higher
        use_reverse : bool, default=False
            Whether to include reversed symptoms (e.g., 'not X'); this will \
            double the size of the feature space.
        top_n : int, default=100
            Number of top-performing combinations to save.
        batch_keep_n : int, default=15
            Number of top combos to keep from each batch. Only the top \
            combinations will be kept as candidates for the final cut. Usually \
            only applies for compound combinations.
        write_full : bool, default=True
            Whether to write the full (pre-pruned) set of results to disk. If 
            prune is on, only the pruned results will be saved when this is off.
        prune : bool, default=True
            Whether to discard combinatinos from each batch that are worse than 
            the best result from previous batches.
        tol : float, default=0.05,
            Maximum amount of a metric a combination in a batch can lag the max 
            from previous batches before being discarded. Only applies when 
            prune is set to True. 
        
        Returns
        ----------
        
        """
        # Separating the column names and values
        symptom_list = X.columns.values
        self.var_names = symptom_list
        X = X.values.astype(np.uint8)
        y = y.astype(np.uint8)
        self.X = X
        self.y = y
                
        # Setting up a lookup dictionary for the metric names
        self.metric = metric
        fn1 = metric_dict[metric]['fn1']
        fn2 = metric_dict[metric]['fn2']
        
        # Optionally adding reversed versions of the columns
        if use_reverse:
            X_rev = -X + 1
            X = np.concatenate((X, X_rev), axis=1)
        
        # Setting max_n, if not already specified
        if not max_n:
            max_n = n_symp
        
        # Optionally copying the data into shared memory
        if self.share_memory:
            shm_x = shared_memory.SharedMemory(name='predictors',
                                               create=True, 
                                               size=X.nbytes)
            shm_y = shared_memory.SharedMemory(name='outcome',
                                               create=True,
                                               size=y.nbytes)
            X_sh = np.ndarray(X.shape, 
                              dtype=X.dtype, 
                              buffer=shm_x.buf)
            y_sh = np.ndarray(y.shape,
                              dtype=y.dtype,
                              buffer=shm_y.buf)
            X_sh[:] = X[:]
            y_sh[:] = y[:]
            
        # Setting up the combinations
        n_symp = X.shape[1]
        n_vals = list(range(1, max_n + 1))
        
        # Dropping impossible symptom pairings
        """
        if use_reverse:
            clashes = [[i, i + n_symp] for i in range(n_symp)]
            keepers = [[np.sum([c[0] in l and c[1] in l
                                for c in clashes]) == 0
                        for l in combos]
                     for combos in col_combos]
            col_combos = [[c for j, c in enumerate(combos) if keepers[i][j]]
                        for i, combos in enumerate(col_combos)]
            symptom_list += ['no_' + s for s in symptom_list]
        """
        
        # Running the simple search loop
        symp_out = []
        if verbose:
            print('Evaluating the simple combinations.')
        
        with Pool(processes=self.n_jobs) as p:
            for n_val in n_vals:
                for m in range(n_val):
                    c_out = []
                    combo_iter = combinations(range(n_symp), n_val)
                    combos_remaining = True
                    while combos_remaining:
                        combos = list(islice(combo_iter, chunksize))
                        if len(combos) == 0:
                            combos_remaining = False
                        else:
                            if self.share_memory:
                                inputs = [(c, m, X.shape, metric) 
                                          for c in combos]
                                res = p.starmap(metrics.shared_mem_score, 
                                                inputs)
                            else:
                                slices = [np.array(X[:, c].sum(1) > m, 
                                                   dtype=np.uint8) 
                                          for c in combos]
                                inputs = [(y, s, metric) for s in slices]
                                res = p.starmap(metrics.score_set, inputs)
                            res = np.array(res)
                            combo_names = [' '.join([symptom_list[i]
                                                     for i in np.array(c)])
                                           for c in combos]
                            mn = np.array([[m + 1, n_val]] * res.shape[0])
                            cols = np.array(combo_names).reshape(-1, 1)
                            c_out.append(np.concatenate([mn, cols, res], 
                                                        axis=1))
                    symp_out.append(np.concatenate(c_out, axis=0))
            p.close()
            p.join()
        
        if self.share_memory:
            self.clear_shared_memory()
        
        results = pd.DataFrame(np.concatenate(symp_out),
                               columns=['m1', 'n1', 'rule1',
                                         fn1, fn2, metric])
        results.sort_values(metric,
                            ascending=False,
                            inplace=True)
        results[['m2', 'n2']] = 0
        results[['rule2', 'link']] = ''
        results = results[['m1', 'rule1', 'link',
                           'm2', 'rule2', 'n1',
                           'n2', fn1, fn2, metric]]
        float_cols = [fn1, fn2, metric]
        int_cols = ['m1', 'm2', 'n1', 'n2']
        results[float_cols] = results[float_cols].astype(float)
        results[int_cols] = results[int_cols].astype(int)

        if top_n:
            results = results.iloc[:top_n, :]
            results.reset_index(inplace=True, drop=True)

        if write_full:
            results.to_csv('data/' + csv_name, index=False)

        self.results = results

        if compound:
            # Make the initial list of combinations of the column combinations
            col_combos = [list(combinations(range(n_symp), i)) 
                        for i in range(1, max_n + 1)]
            col_combos = tools.flatten(col_combos)
            meta_iter = combinations(col_combos, 2)
            
            with Pool(processes=self.n_jobs) as p:
                if verbose:
                    print('Building the list of compound combinations.')
                
                metacombos = p.map(tools.unique_combo, meta_iter)
                p.close()
                p.join()
            
            metacombos = [c for c in metacombos if c is not None]
            
            # Combos of m for m-of-n
            mcs = [pair for pair in permutations(range(1, max_n + 1), 2)]
            mcs += [(i, i) for i in range(1, max_n + 1)]
            
            # Sums for at least 2 
            raw_pairsums = [[(X, c, mc, y, metric) 
                             for c in metacombos] for mc in mcs]
            
            # Weeding out combinations that don't make sense
            pairsum_input = []
            for i, pairs in enumerate(raw_pairsums):
                good_pairs = [pair for pair in pairs
                              if len(pair[1][0]) >= pair[2][0] 
                              and len(pair[1][1]) >= pair[2][1]]
                pairsum_input.append(good_pairs)
            pairsum_input = [input for input in pairsum_input
                             if len(input) > 0]
            
            # Max number of combos to consider from 'and', 'or', and 'any'
            if verbose:
                print('Running the evaluation loop.')
            
            num_runs = str(len(pairsum_input))
            with Pool(processes=self.n_jobs) as p:
                for run_num, input in enumerate(pairsum_input):
                    batch_n = np.min([len(input), batch_keep_n])
                    if verbose:
                        batch_mess = str(run_num + 1) + ' of ' + num_runs
                        print('')
                        print('Running batch ' + batch_mess)
                    
                    if progress_bars:
                        input = tqdm.tqdm(input)
                    
                    # Calculating f1 score for each of the combo sums
                    scores = p.starmap(metrics.pair_score, input)
                    and_scores = pd.DataFrame([s[0] for s in scores])
                    or_scores = pd.DataFrame([s[1] for s in scores])
                    all_scores = pd.concat([and_scores, or_scores], axis=0)
                    all_scores.columns = [fn1, fn2, metric]
                    
                    # Getting the names of the rules as strings
                    or_input = [(pair, 'or', self.var_names) 
                                 for pair in input]
                    or_info = p.starmap(tools.pair_info, or_input)
                    or_df = pd.DataFrame(or_info,
                                         columns=['m1', 'rule1', 'link',
                                                  'm2', 'rule2', 'n1', 'n2'])
                    and_df = deepcopy(or_df)
                    and_df['link'] = 'and'
                    info_df = pd.concat([and_df, or_df], axis=0)
                    info_df = pd.concat([info_df, all_scores], axis=1)
                    batch_res = info_df.sort_values(metric,
                                                    ascending=False)
                    batch_res = batch_res.iloc[0:batch_n, :]
                    
                    if write_full:
                        batch_res.to_csv('data/' + csv_name,
                                         mode='a',
                                         header=False,
                                         index=False)
                    if prune:
                        best_max = self.results[metric].max()
                        current_max = batch_res[metric].max()
                        if (current_max - best_max) >= -tol:
                            self.results = pd.concat([batch_res,
                                                      self.results], axis=0)
                        else:
                            print('Previous max was ' + str(best_max))
                            print('Current max is only ' + str(current_max))
                            print('Discarding current batch of rules.')
                    else:
                        self.results = pd.concat([batch_res,
                                                  self.results], aixs=0)

            p.close()
            p.join()
        
        self.results.sort_values(metric, ascending=False, inplace=True)
        self.results.reset_index(inplace=True, drop=True)
        self.results['total_n'] = self.results.n1 + self.results.n2
        
        return
    
    def clear_shared_memory(self):
        shm_x = shared_memory.SharedMemory(name='predictors',
                                           create=False, 
                                           size=self.X.nbytes)
        shm_y = shared_memory.SharedMemory(name='outcome',
                                           create=False,
                                           size=self.y.nbytes)
        shm_x.close()
        shm_x.unlink()
        shm_y.close()
        shm_y.unlink()
    
    def plot(self, 
             metric=None, 
             mark_best=True,
             separate_n=False,
             hue=None,
             grid_style='darkgrid',
             palette='crest',
             font_scale=1,
             add_hull=False):
        """Plots combinations in the selected metric's space."""
        sns.set_style('darkgrid')
        sns.set(font_scale=font_scale)
        cp = sns.color_palette(palette)
        
        if not metric:
            metric = self.metric
        
        if not hue:
            hue = metric
        
        md = metric_dict[self.metric]
        fn1, fn2 = md['fn1'], md['fn2']
        xlab, ylab = md['xlab'], md['ylab']
        
        y = self.results[fn1].values
        x = self.results[fn2].values
        
        if metric == 'j':
            x = 1 - self.results[fn2].values
        
        rp_col = 'total_n' if separate_n else None
        
        rp = sns.relplot(x=x, 
                         y=y, 
                         hue=self.results[metric], 
                         col=rp_col, 
                         data=self.results,
                         kind='scatter',
                         palette=palette)
        
        if mark_best and not separate_n:
            plt.scatter(x[0], y[0], 
                        color='black',
                        marker='x')
            
        rp.set(xlim=(0, 1), ylim=(0, 1))
        rp.fig.set_tight_layout(True)
        rp.set_xlabels(xlab)
        rp.set_ylabels(ylab)        
        
        plt.show()
        
        
    def predict(self, X):
        return tools.rule_to_y(X, self.results.loc[0])


class NonlinearApproximation:
    """A nonlinear approximation to the integer program. Playing it fast and 
    loose.
    """
    def __int__(self):
        return
    
    def fit(self, X, y,
            compound=False,
            num_bags=2,
            max_m=None,
            max_n=None,
            min_sens=None,
            min_spec=None):
        """Fits the optimizer.
        
        Parameters
        ----------
          max_n : int, default=5
            The maximum allowable combination size.
          metric : str, default='j'
            The classification metric to be optimized.
          metric_mode : str, default='max'
            Whether to minimize ('min') or maximimze ('max') the metric.
          compound : bool, default=False
            Whether to search compound combinations. Performance will \
            probably be higher
          use_reverse: bool, default=False
            Whether to include reversed symptoms (e.g., 'not X'); this will \
            double the size of the feature space.
          n_jobs : int, default=-1
            Number of jobs for multiprocessing. -1 runs them all.
          top_n : int, default=100
            Number of top-performing combinations to save.
          batch_keep_n: int, default=15
            Number of top combos to keep from each batch. Only the top \
            combinations will be kept as candidates for the final cut. Usually \
            only applies for compound combinations.
        """
        # Setting up the problem
        self.var_names = X.columns.values
        X = X.values
        
        pos = np.where(y == 1)[0]
        neg = np.where(y == 0)[0]
        
        X = np.concatenate([X[pos], X[neg]], axis=0)
        y = np.concatenate([y[pos], y[neg]], axis=0)
        
        xp = X[:len(pos), :]
        xn = X[len(pos):, :]
        
        N = X.shape[0]
        Ns = X.shape[1]
        bnds = ((0, 1),) * Ns
        bnds += ((1, Ns),)
        init = np.zeros(Ns + 1)
        
        if not compound:
            # Setting up the optional constraints
            cons = []
            if max_n:
                n = 4
                nA = np.concatenate([np.ones(Ns), np.zeros(1)])
                cons.append(sp.optimize.LinearConstraint(nA, lb=1, ub=n))
            if max_m:
                m = 1
                mA = np.concatenate([np.zeros(Ns), np.ones(1)])
                cons.append(sp.optimize.LinearConstraint(mA, lb=1, ub=m))
            if min_sens:
                cons.append(sp.optimize.NonlinearConstraint(metrics.sens_exp,
                                                            lb=min_sens,
                                                            ub=1.0))
            if min_spec:
                cons.append(sp.optimize.NonlinearConstraint(metrics.spec_exp,
                                                            lb=0.8,
                                                            ub=1.0))
            # Running the program
            opt = sp.optimize.minimize(
                fun=metrics.j_exp,
                x0=init,
                args=(xp, xn),
                constraints=cons,
                bounds=bnds
            )
            self.opt = opt
            rounded = opt.x.round()
            self.z = rounded[:-1]
            self.m = rounded[-1]
            
            # Calculating the result's scores
            rule_cols = tools.zm_to_rule(self.z, self.m, self.var_names)
            y_ = self.predict(X)
            score_cols = metrics.score_set(y, y_, 'j', True)
            self.results = pd.concat([rule_cols, score_cols], axis=1)
            return
        else:
            # Now trying the compound program
            Nc = num_bags
            z_bnds = ((0, 1),) * Ns * Nc
            m_bnds = ((0, Ns),) * Nc
            bnds = z_bnds + m_bnds
            
            def m_morethan_n(z):
                m = z[-Nc:]
                z = z[:-Nc]
                z = z.reshape((Ns, Nc), order='F')
                z = tools.smash_log(z - .5, B=15)
                nvals = z.sum(0)
                diffs = tools.smash_log(nvals - m + .5)
                return (Nc - diffs.sum()) 
            
            # Constraint so that no symptom appears in more than one combo
            z_con_mat = np.concatenate([np.identity(Ns)] * Nc, axis=1)
            m_con_mat = np.zeros((Ns, Nc))
            nmax_mat = np.concatenate([z_con_mat, m_con_mat], axis=1)
            nmax_cons = sp.optimize.LinearConstraint(nmax_mat, lb=0, ub=1)
            
            # Constraint so that m <= n for any combo
            z_c_rows = [np.ones(Ns)] * Nc
            z_c_mat = np.zeros((Nc, Ns * Nc))
            for i, r in enumerate(z_c_rows):
                start = i * Ns
                end = start + Ns
                z_c_mat[i, start:end] = r
            
            z_c_mat = np.concatenate([z_c_mat, np.identity(Nc) * -1],
                                     axis=1)
            mn_cons = sp.optimize.LinearConstraint(z_c_mat, 
                                                   lb=0, 
                                                   ub=np.inf)
            mn_cons = sp.optimize.NonlinearConstraint(m_morethan_n, 
                                                      lb=-np.inf, 
                                                      ub=0.999)
            
            # Constraint that at least one combo must have m >= 1
            m_sum = np.concatenate([np.zeros(Ns * Nc),
                                    np.ones(Nc)])
            m_sum_cons = sp.optimize.LinearConstraint(m_sum, lb=1, ub=np.inf)
            
            # Running the optimization
            init = np.zeros(len(bnds))
            opt = sp.optimize.minimize(
                fun=metrics.j_exp_comp,
                x0=init,
                args=(xp, xn, Nc),
                bounds=bnds,
                method='trust-constr',
                constraints=[nmax_cons, m_sum_cons, mn_cons]
            )
            self.opt = opt
            self.solution = opt.x.round()
            self.m = self.solution[-Nc:]
            self.z = self.solution[:-Nc].reshape((Ns, Nc), order='F')
            return metrics.j_lin_comp(self.z, self.m, X, y)
    
    def predict(self, X):
        """Returns the optimal guesses for a new set of data."""
        return tools.zm_to_y(self.z, self.m, X)


class IntegerProgram:
    """The integer program. Always optimal, and usually fast."""
    def __init__(self):
        return
    
    def fit(self, X, y,
            max_n=None,
            solver='CP-SAT'):
        """Fits the optimizer.
        
        Parameters
        ----------
          max_n : int, default=5
            The maximum allowable combination size.
          solver : str, default='CP-SAT'
            Which ortools solver to use. 'SCIP' will also work, but can be 
            really slow.
          
        """
        self.var_names = X.columns.values
        X = X.values
        
        pos = np.where(y == 1)[0]
        neg = np.where(y == 0)[0]
        
        X = np.concatenate([X[pos], X[neg]], axis=0)
        y = np.concatenate([y[pos], y[neg]], axis=0)
        
        xp = X[:len(pos), :]
        xn = X[len(pos):, :]
        
        N = X.shape[0]
        Ns = X.shape[1]
        H = 100
        
        npos = len(pos)
        nneg = len(neg)
        
        T_con = np.identity(N) * H * -1
        m_con_pos = np.ones((npos, 1))
        m_con_neg = np.ones((nneg, 1)) * -1
        m_con = np.concatenate([m_con_pos, m_con_neg])
        Z_con = np.concatenate([xp * -1, xn], axis=0)
        mn_con = np.concatenate([np.ones(Ns) * -1, 
                                 np.ones(1), 
                                 np.zeros(N)]).reshape(1, -1)
        con_mat = np.concatenate([Z_con, m_con, T_con], axis=1)
        con_mat = np.concatenate([con_mat, mn_con], axis=0).astype(np.int16)
        
        # Setting up the bounds on the variables
        bnds = [(0, 1)] * Ns
        bnds += [(1, Ns)]
        bnds += [(0, 1)] * N
        
        # Setting up the bounds on the constraints
        con_bounds = np.concatenate([np.zeros(npos), 
                                     np.ones(nneg) * -1,
                                     np.zeros(1)])
        
        # And setting up the objective
        divs = [1 / (xp.shape[0] / X.shape[0])] * xp.shape[0]
        divs += [1 / (xn.shape[0] / X.shape[0])] * xn.shape[0]
        obj = np.concatenate([np.zeros(Ns),
                              np.zeros(1),
                              np.array(divs)])
        
        def create_data_model():
            """Stores the data for the problem."""
            data = {}
            data['constraint_coeffs'] = [[int(i) for i in j]for j in con_mat]
            data['bounds'] = [i for i in con_bounds]
            data['obj_coeffs'] = [int(i) for i in obj]
            data['num_vars'] = con_mat.shape[1]
            data['num_constraints'] = con_mat.shape[0]
            return data
        
        data = create_data_model()
        
        # Create the mip solver with the SCIP backend.
        solver = pywraplp.Solver.CreateSolver('CP-SAT')
        x = {}
        for j in range(data['num_vars']):
            x[j] = solver.IntVar(bnds[j][0], bnds[j][1], '')
        
        neg_inf = -solver.infinity()
        for i in range(data['num_constraints']):
            constraint = solver.RowConstraint(neg_inf, data['bounds'][i], '')
            for j in range(data['num_vars']):
                constraint.SetCoefficient(x[j], data['constraint_coeffs'][i][j])
        
        objective = solver.Objective()
        for j in range(data['num_vars']):
            objective.SetCoefficient(x[j], data['obj_coeffs'][j])
        
        objective.SetMinimization()
        
        status = solver.Solve()
        
        if status == pywraplp.Solver.OPTIMAL:
            print('Objective value =', solver.Objective().Value())
            print('Problem solved in %f milliseconds' % solver.wall_time())
            print('Problem solved in %d iterations' % solver.iterations())
            print('Problem solved in %d branch-and-bound nodes' % solver.nodes())
        else:
            print('The problem does not have an optimal solution.')
        
        z = np.round([x[i].solution_value()
                      for i in range(Ns)]).astype(np.int8)
        m = np.round(x[Ns].solution_value()).astype(np.int8)
        
        self.vars = x
        self.solver = solver
        self.objective = objective
        self.constraint = constraint
        self.z = z
        self.m = m
        rule_cols = tools.zm_to_rule(z, m, self.var_names)
        y_ = self.predict(X)
        score_cols = metrics.score_set(y, y_, 'j', True)
        self.results = pd.concat([rule_cols, score_cols], axis=1)
    
    def predict(self, X):
        """Returns the optimal guesses for a new set of data."""
        return tools.zm_to_y(self.z, self.m, X)
    
                